﻿using System;
using System.Collections.Generic;
using System.Net.Http;
using System.Threading.Tasks;
using Wisej.Web;

namespace Wisej.AzureSSO
{
	/// <summary>
	/// Handles OAuth2 authentication with Azure and interacts with Microsoft Graph API.
	/// </summary>
	public class AzureService
	{
		#region Fields

		private readonly string _clientId;
		private readonly string _tenantId;
		private readonly string _clientSecret;

		private static readonly string Authority = "https://login.microsoftonline.com";
		private static readonly string TokenEndpoint = "/oauth2/v2.0/token";
		private static readonly string GraphApiUrl = "https://graph.microsoft.com/v1.0";

		private readonly HttpClient _httpClient = new();
		private DateTime? _accessTokenExpiration;
		private string _accessToken;
		private dynamic _user;

		private TaskCompletionSource<string> _authorizationCodeTcs;
		private string _hash;

		#endregion

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="AzureService"/> class.
		/// </summary>
		public AzureService(string clientId, string tenantId, string clientSecret)
		{
			_clientId = clientId;
			_tenantId = tenantId;
			_clientSecret = clientSecret;
		}

		#endregion

		#region Public Methods

		/// <summary>
		/// Restores the user session using browser session storage.
		/// </summary>
		public async Task<bool> RestoreSessionAsync()
		{
			var token = await Application.Browser.SessionStorage.GetValueAsync<string>("accessToken");
			var expiration = await Application.Browser.SessionStorage.GetValueAsync<string>("expiration");

			if (string.IsNullOrWhiteSpace(token) || string.IsNullOrWhiteSpace(expiration))
				return false;

			_accessToken = token;
			_accessTokenExpiration = DateTime.Parse(expiration);

			_httpClient.DefaultRequestHeaders.Remove("Authorization");
			_httpClient.DefaultRequestHeaders.Add("Authorization", $"Bearer {token}");

			return _accessTokenExpiration > DateTime.Now;
		}

		/// <summary>
		/// Clears the stored session and token data.
		/// </summary>
		public void ClearSession()
		{
			Application.Browser.SessionStorage.Clear();

			_accessTokenExpiration = null;
			_accessToken = null;
			_user = null;
		}

		/// <summary>
		/// Retrieves the current Azure AD user information.
		/// </summary>
		public async Task<dynamic> GetCurrentUserAsync()
		{
			if (_accessTokenExpiration < DateTime.Now)
				_user = null;

			if (_user != null)
				return _user;

			try
			{
				await ValidateAccessTokenAsync();

				var response = await _httpClient.GetAsync($"{GraphApiUrl}/me");
				response.EnsureSuccessStatusCode();

				var json = await response.Content.ReadAsStringAsync();
				_user = JSON.Parse(json);
			}
			catch
			{
				// Fail quietly or consider logging.
				_user = null;
			}

			return _user;
		}

		#endregion

		#region Private Methods

		/// <summary>
		/// Validates and refreshes the access token if needed.
		/// </summary>
		private async Task<string> ValidateAccessTokenAsync()
		{
			if (!string.IsNullOrEmpty(_accessToken) && _accessTokenExpiration > DateTime.Now)
				return _accessToken;

			var code = await GetAuthorizationCodeAsync();
			var redirectUri = Application.Url;

			var requestData = new FormUrlEncodedContent(new[]
			{
				new KeyValuePair<string, string>("client_id", _clientId),
				new KeyValuePair<string, string>("scope", "https://graph.microsoft.com/User.Read"),
				new KeyValuePair<string, string>("code", code),
				new KeyValuePair<string, string>("client_secret", _clientSecret),
				new KeyValuePair<string, string>("redirect_uri", redirectUri),
				new KeyValuePair<string, string>("grant_type", "authorization_code")
			});

			using var client = new HttpClient();
			var response = await client.PostAsync($"{Authority}/{_tenantId}{TokenEndpoint}", requestData);

			var json = await response.Content.ReadAsStringAsync();
			response.EnsureSuccessStatusCode();

			var tokenResponse = JSON.Parse(json);
			var token = tokenResponse.access_token;
			var expiresIn = tokenResponse.expires_in;

			var expiration = DateTime.Now.AddSeconds((int)expiresIn);

			_httpClient.DefaultRequestHeaders.Remove("Authorization");
			_httpClient.DefaultRequestHeaders.Add("Authorization", $"Bearer {token}");

			SaveAccessToken(token, expiration);
			return _accessToken;
		}

		/// <summary>
		/// Saves the access token and its expiration to memory and browser session storage.
		/// </summary>
		private void SaveAccessToken(string token, DateTime expiration)
		{
			_accessToken = token;
			_accessTokenExpiration = expiration;

			Application.Browser.SessionStorage.SetValue("accessToken", token);
			Application.Browser.SessionStorage.SetValue("expiration", expiration.ToJSON());
		}

		/// <summary>
		/// Triggers the OAuth2 authorization flow and returns the authorization code.
		/// </summary>
		private async Task<string> GetAuthorizationCodeAsync()
		{
			_authorizationCodeTcs = new TaskCompletionSource<string>();
			Application.ApplicationRefresh += Application_ApplicationRefresh;

			_hash = Application.Hash;
			var redirectUri = Application.Url;

			var url = $"{Authority}/{_tenantId}/oauth2/v2.0/authorize" +
					  $"?client_id={_clientId}" +
					  "&response_type=code" +
					  "&response_mode=query" +
					  "&scope=openid profile email https://graph.microsoft.com/User.Read" +
					  $"&redirect_uri={Uri.EscapeDataString(redirectUri)}";

			Application.Navigate(url);

			return await _authorizationCodeTcs.Task;
		}

		/// <summary>
		/// Event handler triggered after the OAuth2 authorization redirect.
		/// </summary>
		private void Application_ApplicationRefresh(object sender, EventArgs e)
		{
			Application.ApplicationRefresh -= Application_ApplicationRefresh;

			var code = Application.QueryString["code"];

			if (!string.IsNullOrEmpty(code))
			{
				_authorizationCodeTcs.SetResult(code);
				Application.Navigate($"/#{_hash}");
			}
			else
			{
				_authorizationCodeTcs.SetException(new Exception("Authorization code not found."));
			}
		}

		#endregion
	}
}
